import six
import functools

from chef.api import ChefAPI, autoconfigure
from chef.environment import Environment
from chef.exceptions import ChefError, ChefAPIVersionError
from chef.search import Search

try:
    from fabric.api import env, task, roles, output
except ImportError as e:
    env = {}
    task = lambda *args, **kwargs: lambda fn: fn
    roles = task

__all__ = ["chef_roledefs", "chef_environment", "chef_query", "chef_tags"]

# Default environment name
DEFAULT_ENVIRONMENT = "_default"

# Default hostname attributes
DEFAULT_HOSTNAME_ATTR = ["cloud.public_hostname", "fqdn"]

# Sentinel object to trigger defered lookup
_default_environment = object()


def _api(api):
    api = api or ChefAPI.get_global() or autoconfigure()
    if not api:
        raise ChefError("Unable to load Chef API configuration")
    return api


class Roledef(object):
    """Represents a Fabric roledef for a Chef role."""

    def __init__(self, query, api, hostname_attr, environment=None):
        self.query = query
        self.api = api
        self.hostname_attr = hostname_attr
        if isinstance(self.hostname_attr, six.string_types):
            self.hostname_attr = (self.hostname_attr,)
        self.environment = environment

    def __call__(self):
        query = self.query
        environment = self.environment
        if environment is _default_environment:
            environment = env.get("chef_environment", DEFAULT_ENVIRONMENT)
        if environment:
            query += " AND chef_environment:%s" % environment
        for row in Search("node", query, api=self.api):
            if row:
                if callable(self.hostname_attr):
                    val = self.hostname_attr(row.object)
                    if val:
                        yield val
                else:
                    for attr in self.hostname_attr:
                        try:
                            val = row.object.attributes.get_dotted(attr)
                            if (
                                val
                            ):  # Don't ever give out '' or None, since it will error anyway
                                yield val
                                break
                        except KeyError:
                            pass  # Move on to the next
                    else:
                        raise ChefError(
                            "Cannot find a usable hostname attribute for node %s",
                            row.object,
                        )


def chef_roledefs(
    api=None, hostname_attr=DEFAULT_HOSTNAME_ATTR, environment=_default_environment
):
    """Build a Fabric roledef dictionary from a Chef server.

    Example::

        from fabric.api import env, run, roles
        from chef.fabric import chef_roledefs

        env.roledefs = chef_roledefs()

        @roles('web_app')
        def mytask():
            run('uptime')

    ``hostname_attr`` can either be a string that is the attribute in the chef
    node that holds the hostname or IP to connect to, an array of such keys to
    check in order (the first which exists will be used), or a callable which
    takes a :class:`~chef.Node` and returns the hostname or IP to connect to.

    To refer to a nested attribute, separate the levels with ``'.'`` e.g. ``'ec2.public_hostname'``

    ``environment`` is the Chef :class:`~chef.Environment` name in which to
    search for nodes. If set to ``None``, no environment filter is added. If
    set to a string, it is used verbatim as a filter string. If not passed as
    an argument at all, the value in the Fabric environment dict is used,
    defaulting to ``'_default'``.

    .. note::

        ``environment`` must be set to ``None`` if you are emulating Chef API
        version 0.9 or lower.

    .. versionadded:: 0.1

    .. versionadded:: 0.2
        Support for iterable and callable values for  the``hostname_attr``
        argument, and the ``environment`` argument.
    """
    api = _api(api)
    if api.version_parsed < Environment.api_version_parsed and environment is not None:
        raise ChefAPIVersionError(
            "Environment support requires Chef API 0.10 or greater"
        )
    roledefs = {}
    for row in Search("role", api=api):
        name = row["name"]
        roledefs[name] = Roledef("roles:%s" % name, api, hostname_attr, environment)
    return roledefs


@task(alias=env.get("chef_environment_task_alias", "env"))
def chef_environment(name, api=None):
    """A Fabric task to set the current Chef environment context.

    This task works alongside :func:`~chef.fabric.chef_roledefs` to set the
    Chef environment to be used in future role queries.

    Example::

        from chef.fabric import chef_environment, chef_roledefs
        env.roledefs = chef_roledefs()

    .. code-block:: bash

        $ fab env:production deploy

    The task can be configured slightly via Fabric ``env`` values.

    ``env.chef_environment_task_alias`` sets the task alias, defaulting to "env".
    This value must be set **before** :mod:`chef.fabric` is imported.

    ``env.chef_environment_validate`` sets if :class:`~chef.Environment` names
    should be validated before use. Defaults to True.

    .. versionadded:: 0.2
    """
    if env.get("chef_environment_validate", True):
        api = _api(api)
        chef_env = Environment(name, api=api)
        if not chef_env.exists:
            raise ChefError("Unknown Chef environment: %s" % name)
    env["chef_environment"] = name


def chef_query(
    query,
    api=None,
    hostname_attr=DEFAULT_HOSTNAME_ATTR,
    environment=_default_environment,
):
    """A decorator to use an arbitrary Chef search query to find nodes to execute on.

    This is used like Fabric's ``roles()`` decorator, but accepts a Chef search query.

    Example::

        from chef.fabric import chef_query

        @chef_query('roles:web AND tags:active')
        @task
        def deploy():
            pass

    .. versionadded:: 0.2.1
    """
    api = _api(api)
    if api.version_parsed < Environment.api_version_parsed and environment is not None:
        raise ChefAPIVersionError(
            "Environment support requires Chef API 0.10 or greater"
        )
    rolename = "query_" + query
    env.setdefault("roledefs", {})[rolename] = Roledef(
        query, api, hostname_attr, environment
    )
    return lambda fn: roles(rolename)(fn)


def chef_tags(*tags, **kwargs):
    """A decorator to use Chef node tags to find nodes to execute on.

    This is used like Fabric's ``roles()`` decorator, but accepts a list of tags.

    Example::

        from chef.fabric import chef_tags

        @chef_tags('active', 'migrator')
        @task
        def migrate():
            pass

    .. versionadded:: 0.2.1
    """
    # Allow passing a single iterable
    if len(tags) == 1 and not isinstance(tags[0], six.string_types):
        tags = tags[0]
    query = " AND ".join("tags:%s" % tag.strip() for tag in tags)
    return chef_query(query, **kwargs)



# Exception hierarchy for chef
# Copyright (c) 2010 Noah Kantrowitz <noah@coderanger.net>


class ChefError(Exception):
    """Top-level Chef error."""


class ChefServerError(ChefError):
    """An error from a Chef server. May include a HTTP response code."""

    def __init__(self, message, code=None):
        self.raw_message = message
        if isinstance(message, list):
            message = ", ".join(m for m in message if m)
        super(ChefError, self).__init__(message)
        self.code = code

    @staticmethod
    def from_error(message, code=None):
        cls = {
            404: ChefServerNotFoundError,
        }.get(code, ChefServerError)
        return cls(message, code)


class ChefServerNotFoundError(ChefServerError):
    """A 404 Not Found server error."""


class ChefAPIVersionError(ChefError):
    """An incompatible API version error"""


class ChefObjectTypeError(ChefError):
    """An invalid object type error"""



import os

from unittest import mock
import unittest

from chef.api import ChefAPI


class APITestCase(unittest.TestCase):
    def load(self, path):
        path = os.path.join(os.path.dirname(__file__), "configs", path)
        return ChefAPI.from_config_file(path)

    @mock.patch("chef.api.subprocess.Popen")
    def test_config_with_interpolated_settings(self, mock_subproc_popen):
        process_mock = mock.Mock()
        output = (
            b'{"chef_server_url": "http:///chef:4000", "client_key": "../client.pem",'
            b'"node_name": "test_1"}'
        )
        attrs = {"communicate.return_value": (output, "error"), "returncode": 0}
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        api = self.load("basic_with_interpolated_values.rb")
        self.assertEqual(api.client, "test_1")

    def test_basic(self):
        api = self.load("basic.rb")
        self.assertEqual(api.url, "http://chef:4000")
        self.assertEqual(api.client, "test_1")

    def test_current_dir(self):
        api = self.load("current_dir.rb")
        path = os.path.join(os.path.dirname(__file__), "configs", "test_1")
        self.assertEqual(os.path.normpath(api.client), path)

    def test_env_variables(self):
        try:
            os.environ["_PYCHEF_TEST_"] = "foobar"
            api = self.load("env_values.rb")
            self.assertEqual(api.client, "foobar")
        finally:
            del os.environ["_PYCHEF_TEST_"]

    def test_bad_key_raises(self):
        invalids = [None, ""]
        for item in invalids:
            self.assertRaises(ValueError, ChefAPI, "foobar", item, "user")



from chef.base import ChefObject


class Environment(ChefObject):
    """A Chef environment object.

    .. versionadded:: 0.2
    """

    url = "/environments"

    api_version = "0.10"

    attributes = {
        "description": str,
        "cookbook_versions": dict,
        "default_attributes": dict,
        "override_attributes": dict,
    }



from chef import Role
from chef.exceptions import ChefError
from chef.tests import ChefTestCase


class RoleTestCase(ChefTestCase):
    def test_get(self):
        r = Role("test_1")
        self.assertTrue(r.exists)
        self.assertEqual(r.description, "Static test role 1")
        self.assertEqual(r.run_list, [])
        self.assertEqual(r.default_attributes["test_attr"], "default")
        self.assertEqual(r.default_attributes["nested"]["nested_attr"], 1)
        self.assertEqual(r.override_attributes["test_attr"], "override")

    def test_create(self):
        name = self.random()
        r = Role.create(
            name,
            description="A test role",
            run_list=["recipe[foo]"],
            default_attributes={"attr": "foo"},
            override_attributes={"attr": "bar"},
        )
        self.register(r)
        self.assertEqual(r.description, "A test role")
        self.assertEqual(r.run_list, ["recipe[foo]"])
        self.assertEqual(r.default_attributes["attr"], "foo")
        self.assertEqual(r.override_attributes["attr"], "bar")

        r2 = Role(name)
        self.assertTrue(r2.exists)
        self.assertEqual(r2.description, "A test role")
        self.assertEqual(r2.run_list, ["recipe[foo]"])
        self.assertEqual(r2.default_attributes["attr"], "foo")
        self.assertEqual(r2.override_attributes["attr"], "bar")

    def test_delete(self):
        name = self.random()
        r = Role.create(name)
        r.delete()
        for n in Role.list():
            self.assertNotEqual(n, name)
        self.assertFalse(Role(name).exists)



import six.moves
import os
import random
from functools import wraps

from unittest import mock, TestCase, skipUnless

from chef.api import ChefAPI
from chef.exceptions import ChefError
from chef.search import Search

TEST_ROOT = os.path.dirname(os.path.abspath(__file__))


def skipSlowTest():
    return skipUnless(
        os.environ.get("PYCHEF_SLOW_TESTS"),
        "slow tests skipped, set $PYCHEF_SLOW_TESTS=1 to enable",
    )


class mockSearch(object):
    def __init__(self, search_data):
        self.search_data = search_data

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(inner_self):
            return mock.patch("chef.search.Search", side_effect=self._search_inst)(fn)(
                inner_self
            )

        return wrapper

    def _search_inst(self, index, q="*:*", *args, **kwargs):
        data = self.search_data[index, q]
        if not isinstance(data, dict):
            data = {"total": len(data), "rows": data}
        search = Search(index, q, *args, **kwargs)
        search._data = data
        return search


def test_chef_api(**kwargs):
    return ChefAPI(
        "https://api.opscode.com/organizations/pycheftest",
        os.path.join(TEST_ROOT, "client.pem"),
        "unittests",
        **kwargs
    )


class ChefTestCase(TestCase):
    """Base class for Chef unittests."""

    def setUp(self):
        super(ChefTestCase, self).setUp()
        self.api = test_chef_api()
        self.api.set_default()
        self.objects = []

    def tearDown(self):
        for obj in self.objects:
            try:
                obj.delete()
            except ChefError as e:
                print(e)
                # Continue running

    def register(self, obj):
        self.objects.append(obj)

    def random(self, length=8, alphabet="0123456789abcdef"):
        return "".join(random.choice(alphabet) for _ in six.moves.range(length))



from __future__ import absolute_import
import types

try:
    import json
except ImportError:
    import simplejson as json


def maybe_call(x):
    if callable(x):
        return x()
    return x


class JSONEncoder(json.JSONEncoder):
    """Custom encoder to allow arbitrary classes."""

    def default(self, obj):
        if hasattr(obj, "to_dict"):
            return maybe_call(obj.to_dict)
        elif hasattr(obj, "to_list"):
            return maybe_call(obj.to_list)
        elif isinstance(obj, types.GeneratorType):
            return list(obj)
        return super(JSONEncoder, self).default(obj)


loads = json.loads
dumps = lambda obj, **kwargs: json.dumps(obj, cls=JSONEncoder, **kwargs)



import six
import sys
from ctypes import *
from ctypes.util import find_library

if sys.platform == "win32" or sys.platform == "cygwin":
    _eay = CDLL("libeay32.dll")
else:
    _eay = CDLL(find_library("crypto"))

# unsigned long ERR_get_error(void);
ERR_get_error = _eay.ERR_get_error
ERR_get_error.argtypes = []
ERR_get_error.restype = c_ulong

# void ERR_error_string_n(unsigned long e, char *buf, size_t len);
ERR_error_string_n = _eay.ERR_error_string_n
ERR_error_string_n.argtypes = [c_ulong, c_char_p, c_size_t]
ERR_error_string_n.restype = None


class SSLError(Exception):
    """An error in OpenSSL."""

    def __init__(self, message, *args):
        message = message % args
        err = ERR_get_error()
        if err:
            message += ":"
        while err:
            buf = create_string_buffer(120)
            ERR_error_string_n(err, buf, 120)
            message += "\n%s" % string_at(buf, 119)
            err = ERR_get_error()
        super(SSLError, self).__init__(message)


# BIO *   BIO_new(BIO_METHOD *type);
BIO_new = _eay.BIO_new
BIO_new.argtypes = [c_void_p]
BIO_new.restype = c_void_p

# BIO *BIO_new_mem_buf(void *buf, int len);
BIO_new_mem_buf = _eay.BIO_new_mem_buf
BIO_new_mem_buf.argtypes = [c_void_p, c_int]
BIO_new_mem_buf.restype = c_void_p

# BIO_METHOD *BIO_s_mem(void);
BIO_s_mem = _eay.BIO_s_mem
BIO_s_mem.argtypes = []
BIO_s_mem.restype = c_void_p

# long    BIO_ctrl(BIO *bp,int cmd,long larg,void *parg);
BIO_ctrl = _eay.BIO_ctrl
BIO_ctrl.argtypes = [c_void_p, c_int, c_long, c_void_p]
BIO_ctrl.restype = c_long

# define BIO_CTRL_RESET          1  /* opt - rewind/zero etc */
BIO_CTRL_RESET = 1
##define BIO_CTRL_INFO           3  /* opt - extra tit-bits */
BIO_CTRL_INFO = 3

# define BIO_reset(b)            (int)BIO_ctrl(b,BIO_CTRL_RESET,0,NULL)
def BIO_reset(b):
    return BIO_ctrl(b, BIO_CTRL_RESET, 0, None)


##define BIO_get_mem_data(b,pp)  BIO_ctrl(b,BIO_CTRL_INFO,0,(char *)pp)
def BIO_get_mem_data(b, pp):
    return BIO_ctrl(b, BIO_CTRL_INFO, 0, pp)


# int    BIO_free(BIO *a)
BIO_free = _eay.BIO_free
BIO_free.argtypes = [c_void_p]
BIO_free.restype = c_int


def BIO_free_errcheck(result, func, arguments):
    if result == 0:
        raise SSLError("Unable to free BIO")


BIO_free.errcheck = BIO_free_errcheck

# RSA *PEM_read_bio_RSAPrivateKey(BIO *bp, RSA **x,
#                                        pem_password_cb *cb, void *u);
PEM_read_bio_RSAPrivateKey = _eay.PEM_read_bio_RSAPrivateKey
PEM_read_bio_RSAPrivateKey.argtypes = [c_void_p, c_void_p, c_void_p, c_void_p]
PEM_read_bio_RSAPrivateKey.restype = c_void_p

# RSA *PEM_read_bio_RSAPublicKey(BIO *bp, RSA **x,
#                                        pem_password_cb *cb, void *u);
PEM_read_bio_RSAPublicKey = _eay.PEM_read_bio_RSAPublicKey
PEM_read_bio_RSAPublicKey.argtypes = [c_void_p, c_void_p, c_void_p, c_void_p]
PEM_read_bio_RSAPublicKey.restype = c_void_p

# int PEM_write_bio_RSAPrivateKey(BIO *bp, RSA *x, const EVP_CIPHER *enc,
#                                        unsigned char *kstr, int klen,
#                                        pem_password_cb *cb, void *u);
PEM_write_bio_RSAPrivateKey = _eay.PEM_write_bio_RSAPrivateKey
PEM_write_bio_RSAPrivateKey.argtypes = [
    c_void_p,
    c_void_p,
    c_void_p,
    c_char_p,
    c_int,
    c_void_p,
    c_void_p,
]
PEM_write_bio_RSAPrivateKey.restype = c_int

# int PEM_write_bio_RSAPublicKey(BIO *bp, RSA *x);
PEM_write_bio_RSAPublicKey = _eay.PEM_write_bio_RSAPublicKey
PEM_write_bio_RSAPublicKey.argtypes = [c_void_p, c_void_p]
PEM_write_bio_RSAPublicKey.restype = c_int

# int RSA_private_encrypt(int flen, unsigned char *from,
#    unsigned char *to, RSA *rsa,int padding);
RSA_private_encrypt = _eay.RSA_private_encrypt
RSA_private_encrypt.argtypes = [c_int, c_void_p, c_void_p, c_void_p, c_int]
RSA_private_encrypt.restype = c_int

# int RSA_public_decrypt(int flen, unsigned char *from,
#   unsigned char *to, RSA *rsa, int padding);
RSA_public_decrypt = _eay.RSA_public_decrypt
RSA_public_decrypt.argtypes = [c_int, c_void_p, c_void_p, c_void_p, c_int]
RSA_public_decrypt.restype = c_int

RSA_PKCS1_PADDING = 1
RSA_NO_PADDING = 3

# int RSA_size(const RSA *rsa);
RSA_size = _eay.RSA_size
RSA_size.argtypes = [c_void_p]
RSA_size.restype = c_int

# RSA *RSA_generate_key(int num, unsigned long e,
#    void (*callback)(int,int,void *), void *cb_arg);
RSA_generate_key = _eay.RSA_generate_key
RSA_generate_key.argtypes = [c_int, c_ulong, c_void_p, c_void_p]
RSA_generate_key.restype = c_void_p

##define RSA_F4  0x10001L
RSA_F4 = 0x10001

# void RSA_free(RSA *rsa);
RSA_free = _eay.RSA_free
RSA_free.argtypes = [c_void_p]


class Key(object):
    """An OpenSSL RSA key."""

    def __init__(self, fp=None):
        self.key = None
        self.public = False
        if not fp:
            return
        if isinstance(fp, six.binary_type) and fp.startswith(b"-----"):
            # PEM formatted text
            self.raw = fp
        elif isinstance(fp, six.string_types):
            self.raw = open(fp, "rb").read()
        else:
            self.raw = fp.read()
        self._load_key()

    def _load_key(self):
        if b"\0" in self.raw:
            # Raw string has embedded nulls, treat it as binary data
            buf = create_string_buffer(self.raw, len(self.raw))
        else:
            buf = create_string_buffer(self.raw)

        bio = BIO_new_mem_buf(buf, len(buf))
        try:
            self.key = PEM_read_bio_RSAPrivateKey(bio, 0, 0, 0)
            if not self.key:
                BIO_reset(bio)
                self.public = True
                self.key = PEM_read_bio_RSAPublicKey(bio, 0, 0, 0)
            if not self.key:
                raise SSLError("Unable to load RSA key")
        finally:
            BIO_free(bio)

    @classmethod
    def generate(cls, size=1024, exp=RSA_F4):
        self = cls()
        self.key = RSA_generate_key(size, exp, None, None)
        return self

    def private_encrypt(self, value, padding=RSA_PKCS1_PADDING):
        if self.public:
            raise SSLError("private method cannot be used on a public key")
        if six.PY3 and not isinstance(value, bytes):
            buf = create_string_buffer(value.encode(), len(value))
        else:
            buf = create_string_buffer(value, len(value))
        size = RSA_size(self.key)
        output = create_string_buffer(size)
        ret = RSA_private_encrypt(len(buf), buf, output, self.key, padding)
        if ret <= 0:
            raise SSLError("Unable to encrypt data")
        return output.raw[:ret]

    def public_decrypt(self, value, padding=RSA_PKCS1_PADDING):
        if six.PY3 and not isinstance(value, bytes):
            buf = create_string_buffer(value.encode(), len(value))
        else:
            buf = create_string_buffer(value, len(value))
        size = RSA_size(self.key)
        output = create_string_buffer(size)
        ret = RSA_public_decrypt(len(buf), buf, output, self.key, padding)
        if ret <= 0:
            raise SSLError("Unable to decrypt data")
        if six.PY3 and isinstance(output.raw, bytes):
            return output.raw[:ret].decode()
        else:
            return output.raw[:ret]

    def private_export(self):
        if self.public:
            raise SSLError("private method cannot be used on a public key")
        out = BIO_new(BIO_s_mem())
        PEM_write_bio_RSAPrivateKey(out, self.key, None, None, 0, None, None)
        buf = c_char_p()
        count = BIO_get_mem_data(out, byref(buf))
        pem = string_at(buf, count)
        BIO_free(out)
        return pem

    def public_export(self):
        out = BIO_new(BIO_s_mem())
        PEM_write_bio_RSAPublicKey(out, self.key)
        buf = c_char_p()
        count = BIO_get_mem_data(out, byref(buf))
        pem = string_at(buf, count)
        BIO_free(out)
        return pem

    def __del__(self):
        if self.key and RSA_free:
            RSA_free(self.key)



import unittest

from chef import Client
from chef.tests import ChefTestCase


class ClientTestCase(ChefTestCase):
    def test_list(self):
        self.assertIn("test_1", Client.list())

    def test_get(self):
        client = Client("test_1")
        self.assertTrue(client.platform)
        self.assertEqual(client.orgname, "pycheftest")
        self.assertTrue(client.public_key)
        self.assertTrue(client.certificate)
        self.assertEqual(client.private_key, None)

    def test_create(self):
        name = self.random()
        client = Client.create(name)
        self.register(client)
        self.assertEqual(client.name, name)
        # self.assertEqual(client.orgname, 'pycheftest') # See CHEF-2019
        self.assertTrue(client.private_key)
        self.assertTrue(client.public_key)
        self.assertIn(name, Client.list())

        client2 = Client(name)
        client2.rekey()
        self.assertEqual(client.public_key, client2.public_key)
        self.assertNotEqual(client.private_key, client2.private_key)

    def test_delete(self):
        name = self.random()
        client = Client.create(name)
        client.delete()
        self.assertNotIn(name, Client.list())



import six
import abc
from collections.abc import MutableMapping

from chef.api import ChefAPI
from chef.base import ChefObject, ChefQuery, ChefObjectMeta
from chef.exceptions import ChefError, ChefServerNotFoundError


class DataBagMeta(ChefObjectMeta, abc.ABCMeta):
    """A metaclass to allow DataBag to use multiple inheritance."""


class DataBag(six.with_metaclass(DataBagMeta, ChefObject, ChefQuery)):
    """A Chef data bag object.

    Data bag items are available via the mapping API. Evaluation works in the
    same way as :class:`ChefQuery`, so requesting only the names will not
    cause the items to be loaded::

        bag = DataBag('versions')
        item = bag['web']
        for name, item in six.iteritems(bag):
            print item['qa_version']
    """

    url = "/data"

    def _populate(self, data):
        self.names = list(data.keys())

    def obj_class(self, name, api):
        return DataBagItem(self, name, api=api)


class DataBagItem(
    six.with_metaclass(DataBagMeta, ChefObject, MutableMapping)
):
    """A Chef data bag item object.

    Data bag items act as normal dicts and can contain arbitrary data.
    """

    url = "/data"
    attributes = {
        "raw_data": dict,
    }

    def __init__(self, bag, name, api=None, skip_load=False):
        self._bag = bag
        super(DataBagItem, self).__init__(
            str(bag) + "/" + name, api=api, skip_load=skip_load
        )
        self.name = name

    @property
    def bag(self):
        """The :class:`DataBag` this item is a member of."""
        if not isinstance(self._bag, DataBag):
            self._bag = DataBag(self._bag, api=self.api)
        return self._bag

    @classmethod
    def from_search(cls, data, api):
        bag = data.get("data_bag")
        if not bag:
            raise ChefError("No data_bag key in data bag item information")
        name = data.get("name")
        if not name:
            raise ChefError("No name key in the data bag item information")
        item = name[len("data_bag_item_" + bag + "_") :]
        obj = cls(bag, item, api=api, skip_load=True)
        obj.exists = True
        obj._populate(data)
        return obj

    def _populate(self, data):
        if "json_class" in data:
            self.raw_data = data["raw_data"]
        else:
            self.raw_data = data

    def __len__(self):
        return len(self.raw_data)

    def __iter__(self):
        return iter(self.raw_data)

    def __getitem__(self, key):
        return self.raw_data[key]

    def __setitem__(self, key, value):
        self.raw_data[key] = value

    def __delitem__(self, key):
        del self.raw_data[key]

    @classmethod
    def create(cls, bag, name, api=None, **kwargs):
        """Create a new data bag item. Pass the initial value for any keys as
        keyword arguments."""
        api = api or ChefAPI.get_global()
        obj = cls(bag, name, api, skip_load=True)
        for key, value in six.iteritems(kwargs):
            obj[key] = value
        obj["id"] = name
        api.api_request("POST", cls.url + "/" + str(bag), data=obj.raw_data)
        if isinstance(bag, DataBag) and name not in bag.names:
            # Mutate the bag in-place if possible, so it will return the new
            # item instantly
            bag.names.append(name)
        return obj

    def save(self, api=None):
        """Save this object to the server. If the object does not exist it
        will be created.
        """
        api = api or self.api
        self["id"] = self.name
        try:
            api.api_request("PUT", self.url, data=self.raw_data)
        except ChefServerNotFoundError as e:
            api.api_request(
                "POST", self.__class__.url + "/" + str(self._bag), data=self.raw_data
            )



import six
from collections.abc import MutableMapping, Sequence


from chef.base import ChefObject
from chef.exceptions import ChefError


class NodeAttributes(MutableMapping):
    """A collection of Chef :class:`~chef.Node` attributes.

    Attributes can be accessed like a normal python :class:`dict`::

        print node['fqdn']
        node['apache']['log_dir'] = '/srv/log'

    When writing to new attributes, any dicts required in the hierarchy are
    created automatically.

    .. versionadded:: 0.1
    """

    def __init__(self, search_path=[], path=None, write=None):
        if not isinstance(search_path, Sequence):
            search_path = [search_path]
        self.search_path = search_path
        self.path = path or ()
        self.write = write

    def __iter__(self):
        keys = set()
        for d in self.search_path:
            keys |= set(six.iterkeys(d))
        return iter(keys)

    def __len__(self):
        l = 0
        for key in self:
            l += 1
        return l

    def __getitem__(self, key):
        for d in self.search_path:
            if key in d:
                value = d[key]
                break
        else:
            raise KeyError(key)
        if not isinstance(value, dict):
            return value
        new_search_path = []
        for d in self.search_path:
            new_d = d.get(key, {})
            if not isinstance(new_d, dict):
                # Structural mismatch
                new_d = {}
            new_search_path.append(new_d)
        return self.__class__(new_search_path, self.path + (key,), write=self.write)

    def __setitem__(self, key, value):
        if self.write is None:
            raise ChefError("This attribute is not writable")
        dest = self.write
        for path_key in self.path:
            dest = dest.setdefault(path_key, {})
        dest[key] = value

    def __delitem__(self, key):
        if self.write is None:
            raise ChefError("This attribute is not writable")
        dest = self.write
        for path_key in self.path:
            dest = dest.setdefault(path_key, {})
        del dest[key]

    def has_dotted(self, key):
        """Check if a given dotted key path is present. See :meth:`.get_dotted`
        for more information on dotted paths.

        .. versionadded:: 0.2
        """
        try:
            self.get_dotted(key)
        except KeyError:
            return False
        else:
            return True

    def get_dotted(self, key):
        """Retrieve an attribute using a dotted key path. A dotted path
        is a string of the form `'foo.bar.baz'`, with each `.` separating
        hierarcy levels.

        Example::

            node.attributes['apache']['log_dir'] = '/srv/log'
            print node.attributes.get_dotted('apache.log_dir')
        """
        value = self
        for k in key.split("."):
            if not isinstance(value, NodeAttributes):
                raise KeyError(key)
            value = value[k]
        return value

    def set_dotted(self, key, value):
        """Set an attribute using a dotted key path. See :meth:`.get_dotted`
        for more information on dotted paths.

        Example::

            node.attributes.set_dotted('apache.log_dir', '/srv/log')
        """
        dest = self
        keys = key.split(".")
        last_key = keys.pop()
        for k in keys:
            if k not in dest:
                dest[k] = {}
            dest = dest[k]
            if not isinstance(dest, NodeAttributes):
                raise ChefError
        dest[last_key] = value

    def to_dict(self):
        merged = {}
        for d in reversed(self.search_path):
            merged.update(d)
        return merged


class Node(ChefObject):
    """A Chef node object.

    The Node object can be used as a dict-like object directly, as an alias for
    the :attr:`.attributes` data::

        >>> node = Node('name')
        >>> node['apache']['log_dir']
        '/var/log/apache2'

    .. versionadded:: 0.1

    .. attribute:: attributes

        :class:`~chef.node.NodeAttributes` corresponding to the composite of all
        precedence levels. This only uses the stored data on the Chef server,
        it does not merge in attributes from roles or environments on its own.

        ::

            >>> node.attributes['apache']['log_dir']
            '/var/log/apache2'

    .. attribute:: run_list

        The run list of the node. This is the unexpanded list in ``type[name]``
        format.

        ::

            >>> node.run_list
            ['role[base]', 'role[app]', 'recipe[web]']

    .. attribute:: chef_environment

        The name of the Chef :class:`~chef.Environment` this node is a member
        of. This value will still be present, even if communicating with a Chef
        0.9 server, but will be ignored.

        .. versionadded:: 0.2

    .. attribute:: default

        :class:`~chef.node.NodeAttributes` corresponding to the ``default``
        precedence level.

    .. attribute:: normal

        :class:`~chef.node.NodeAttributes` corresponding to the ``normal``
        precedence level.

    .. attribute:: override

        :class:`~chef.node.NodeAttributes` corresponding to the ``override``
        precedence level.

    .. attribute:: automatic

        :class:`~chef.node.NodeAttributes` corresponding to the ``automatic``
        precedence level.
    """

    url = "/nodes"
    attributes = {
        "default": NodeAttributes,
        "normal": lambda d: NodeAttributes(d, write=d),
        "override": NodeAttributes,
        "automatic": NodeAttributes,
        "run_list": list,
        "chef_environment": str,
    }

    def has_key(self, key):
        return self.attributes.has_dotted(key)

    def get(self, key, default=None):
        return self.attributes.get(key, default)

    def __getitem__(self, key):
        return self.attributes[key]

    def __setitem__(self, key, value):
        self.attributes[key] = value

    def _populate(self, data):
        if not self.exists:
            # Make this exist so the normal<->attributes cross-link will
            # function correctly
            data["normal"] = {}
        data.setdefault("chef_environment", "_default")
        super(Node, self)._populate(data)
        self.attributes = NodeAttributes(
            (
                data.get("automatic", {}),
                data.get("override", {}),
                data["normal"],  # Must exist, see above
                data.get("default", {}),
            ),
            write=data["normal"],
        )

    def cookbooks(self, api=None):
        api = api or self.api
        return api[self.url + "/cookbooks"]



import pkg_resources
from chef.api import ChefAPI
from chef.exceptions import ChefObjectTypeError
from chef.permissions import Permissions


class Acl(object):
    """
    Acl class provides access to the Acl in the Chef 12

        Acl(object_type, name, api, skip_load=False)

        - object_type - type of the Chef object. Can be one of the following value: "clients", "containers", "cookbooks",
            "data", "environments", "groups", "nodes", "roles"
        - name - name of the Chef object (e.g. node name)
        - api - object of the ChefAPI class, configured to work with Chef server
        - skip_load - is skip_load is False, new object will be initialized with current Acl settings of the specified
            object

        Example::

            from chef import ChefAPI
            from chef.acl import Acl

            api = ChefAPI('http://chef.com:4000', 'chef-developer.pem', 'chef-developer', '12.0.0')
            acl = Acl('nodes', 'i-022fcb0d', api)

        Each object of the Acl class contains the following properties:
            create, read, update, delete, grant
        each property represents corresponding access rights to the Chef object.
        each property contains the following fields (https://github.com/astryia/pychef/blob/acls/chef/permissions.py):
        - actors - list of the users, which have corresponding permissions
        - groups - list of the groups, which have corresponding permissions

        Example::

            print acl.update.groups
            >>> ['admins', 'clients']

        Each object of the class Acl contains the following methods:
        - reload() - reload current Acls from the Chef server
        - save() - save updated Acl object to the Chef server
        - is_supported() - return true if current Api version supports work with Acls

        Example::

            from chef import ChefAPI
            from chef.acl import Acl

            api = ChefAPI('http://chef.com:4000', 'chef-developer.pem', 'chef-developer', '12.0.0')
            acl = Acl('nodes', 'i-022fcb0d', api)
            print acl.update.groups
            >>> ['admins']
            acl.update.groups.append('clients')
            acl.save()
            acl.reload()
            print acl.update.groups
            >>> ['admins', 'clients']

        Each class which represents Chef object contains method get_acl() method

        Example::

            from chef import ChefAPI
            from chef.node import Node

            api = ChefAPI('http://chef.com:4000', 'chef-developer.pem', 'chef-developer', '12.0.0')
            node = Node('i-022fcb0d', api)
            acl = node.get_acl()
            print acl.read.groups
            >>> ['admins']
            acl.save()

        Note about versions
        Chef server with version < 12 doesn't have Acl endpoint, so, I've introduced method is_supported() for Acl class.
        This method check if api version is greater than 12.
        So you should pass valid Chef server version to the ChefAPI constructor

        Example::

            api = ChefAPI('http://chef.com:4000', 'chef-developer.pem', 'chef-developer', '12.0.0')
            acl = Acl('nodes', 'i-022fcb0d', api)
            print acl.is_supported()
            >>> True

            api = ChefAPI('http://chef.com:4000', 'chef-developer.pem', 'chef-developer', '11.2.0')
            acl = Acl('nodes', 'i-022fcb0d', api)
            print acl.is_supported()
            >>> False

        But if you pass string '12.0.0' when actual Chef server version is 11.2, you will receive an error when you try
        to build Acl object.
    """

    ace_types = ["create", "read", "update", "delete", "grant"]
    object_types = [
        "clients",
        "containers",
        "cookbooks",
        "data",
        "environments",
        "groups",
        "nodes",
        "roles",
    ]

    """ ALC API available only in Chef server from version 12.0"""
    version = pkg_resources.parse_version("12.0.0")

    def __init__(self, object_type, name, api, skip_load=False):
        self._check_object_type(object_type)

        self.object_type = object_type
        self.name = name
        self.url = "/%s/%s/_acl/" % (object_type, name)
        self.api = api or ChefAPI.get_global()

        self.attributes_map = {}
        for t in self.ace_types:
            self.attributes_map[t] = Permissions()

        if (not skip_load) and self.is_supported():
            self.reload()

    @property
    def create(self):
        """Gets Create permissions"""
        return self.attributes_map["create"]

    @property
    def read(self):
        """Gets Read permissions"""
        return self.attributes_map["read"]

    @property
    def update(self):
        """Gets Update permissions"""
        return self.attributes_map["update"]

    @property
    def delete(self):
        """Gets Delete permissions"""
        return self.attributes_map["delete"]

    @property
    def grant(self):
        """Gets Grant permissions"""
        return self.attributes_map["grant"]

    def save(self):
        """Save updated permissions objects to the Chef server"""
        for t in self.ace_types:
            self.api.api_request("PUT", self.url + t, data={t: self[t]})

    def __getitem__(self, key):
        if key in self.attributes_map.keys():
            return self.attributes_map[key]
        else:
            return {}

    def reload(self):
        """Load current permissions for the object"""
        data = self.api.api_request("GET", self.url)
        for t in self.ace_types:
            self[t].actors = data[t]["actors"]
            self[t].groups = data[t]["groups"]

    def to_dict(self):
        d = {}
        for t in self.ace_types:
            d[t] = self[t].to_dict()

        return d

    def _check_object_type(self, object_type):
        if object_type not in self.object_types:
            raise ChefObjectTypeError("Object type %s is not allowed" % object_type)

    def is_supported(self):
        return self.api.version_parsed >= self.version



from chef.api import ChefAPI
from chef.base import ChefObject


class Client(ChefObject):
    """A Chef client object."""

    url = "/clients"

    def _populate(self, data):
        self.platform = self.api and self.api.platform
        self.private_key = None
        if self.platform:
            self.orgname = data.get("orgname")
            self.validator = bool(data.get("validator", False))
            self.public_key = data.get("public_key")
            self.admin = False
        else:
            self.admin = bool(data.get("admin", False))
            self.public_key = data.get("public_key")
            self.orgname = None
            self.validator = False

    @property
    def certificate(self):
        return self.public_key

    def to_dict(self):
        d = super(Client, self).to_dict()
        d["json_class"] = "Chef::ApiClient"
        if self.platform:
            d.update(
                {
                    "orgname": self.orgname,
                    "validator": self.validator,
                    "public_key": self.certificate,
                    "clientname": self.name,
                }
            )
        else:
            d.update(
                {
                    "admin": self.admin,
                    "public_key": self.public_key,
                }
            )
        return d

    @classmethod
    def create(cls, name, api=None, admin=False):
        api = api or ChefAPI.get_global()
        obj = cls(name, api, skip_load=True)
        obj.admin = admin
        d = api.api_request("POST", cls.url, data=obj)
        obj.private_key = d["private_key"]
        obj.public_key = d["public_key"]
        return obj

    def rekey(self, api=None):
        api = api or self.api
        d_in = {"name": self.name, "private_key": True}
        d_out = api.api_request("PUT", self.url, data=d_in)
        self.private_key = d_out["private_key"]
